Attribute VB_Name = "basPCX"
'==================================================
'PCX decoding functions for Visual Basic!
'
'(C) '99 Ryuunosuke (Dave Hng)
'
'Reads 256 indexed colour, and 24 bit pcx images.
'
'Written with information from ZSoft.
'==================================================

Option Explicit

Public Declare Sub RtlMoveMemory Lib "kernel32" (Destination As Any, Source As Any, ByVal Length As Long)

Public Type FILETIME
    dwLowDateTime As Long
    dwHighDateTime As Long
End Type

Public Type PALETTEENTRY
    peRed As Byte
    peGreen As Byte
    peBlue As Byte
    peFlags As Byte
End Type

Private Type RGBTriple
    Red As Byte
    Green As Byte
    Blue As Byte
End Type

Private Type PCXHeader
    Manufacturer As Byte
    '10 = ZSoft
    
    Version As Byte
    '0 = 2.5
    '2 = 2.8 with palette
    '3 = 2.8 without palette
    '4 = Pc paintbrush for windows
    '5 = Pc Paintbrush ver 3.0+ w. 24 bit encoding
    
    Encoding As Byte
    '1 = .PCX RLE
    
    Bpp As Byte
    '1, 2, 4, 8
    
    XMin As Integer
    YMin As Integer
    XMax As Integer
    YMax As Integer
    
    HDpi As Integer
    VDpi As Integer
    
    ColourPalette(0 To 15) As RGBTriple
    
    Reserved1 As Byte
    
    Planes As Byte
    BytesPerLine As Integer
    PaletteInfo As Integer
    HScreenSize As Integer
    VScreenSize As Integer
    
    Reserved2(0 To 53)  As Byte
End Type

'Bitmap API structs, functions.
'================================================================
Private Const BI_RGB = 0&
Private Const CBM_INIT = &H4
Private Const DIB_RGB_COLORS = 0
Private Const SRCCOPY = &HCC0020
Public Const TrueAPI As Long = 1&

Private Type BITMAPINFOHEADER
    biSize As Long
    biWidth As Long
    biHeight As Long
    biPlanes As Integer
    biBitCount As Integer
    biCompression As Long
    biSizeImage As Long
    biXPelsPerMeter As Long
    biYPelsPerMeter As Long
    biClrUsed As Long
    biClrImportant As Long
End Type

Private Type RGBQUAD
    rgbBlue As Byte
    rgbGreen As Byte
    rgbRed As Byte
    rgbReserved As Byte
End Type

Private Type BITMAPINFO
    bmiHeader As BITMAPINFOHEADER
    bmiColors As RGBQUAD
End Type

Private Type BITMAPINFO_256
    bmiHeader As BITMAPINFOHEADER
    bmiColors(0 To 255) As RGBQUAD
End Type

Private bmpinfo As BITMAPINFO_256
Private hBitmap As Long

Private Declare Function CreateDIBitmap Lib "gdi32" (ByVal hdc As Long, lpInfoHeader As BITMAPINFOHEADER, ByVal dwUsage As Long, lpInitBits As Any, lpInitInfo As BITMAPINFO, ByVal wUsage As Long) As Long
Private Declare Function CreateDIBitmap256 Lib "gdi32" Alias "CreateDIBitmap" (ByVal hdc As Long, lpInfoHeader As BITMAPINFOHEADER, ByVal dwUsage As Long, lpInitBits As Any, lpInitInfo As BITMAPINFO_256, ByVal wUsage As Long) As Long
Private Declare Function GetDC Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function ReleaseDC Lib "user32" (ByVal hWnd As Long, ByVal hdc As Long) As Long
Private Declare Function CreateCompatibleDC Lib "gdi32" (ByVal hdc As Long) As Long
Private Declare Function SelectObject Lib "gdi32" (ByVal hdc As Long, ByVal hObject As Long) As Long
Private Declare Function BitBlt Lib "gdi32" (ByVal hDestDC As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hSrcDC As Long, ByVal xSrc As Long, ByVal ySrc As Long, ByVal dwRop As Long) As Long
Private Declare Function DeleteDC Lib "gdi32" (ByVal hdc As Long) As Long
Private Declare Function DeleteObject Lib "gdi32" (ByVal hObject As Long) As Long
Private Declare Function StretchBlt Lib "gdi32" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hSrcDC As Long, ByVal xSrc As Long, ByVal ySrc As Long, ByVal nSrcWidth As Long, ByVal nSrcHeight As Long, ByVal dwRop As Long) As Long

'Crazy psycho StdPicture creating functions.
'(Courtesy of MS's OLE Automation.. hehe)
'================================================================

Private Declare Function OleCreatePictureIndirect Lib "olepro32.dll" (PicDesc As PicBmp, RefIID As GUID, ByVal fPictureOwnsHandle As Long, IPic As IPicture) As Long

Private Type LOGPALETTE
    palVersion As Integer
    palNumEntries As Integer
    palPalEntry(255) As PALETTEENTRY
End Type

Private Type GUID
    Data1 As Long
    Data2 As Integer
    Data3 As Integer
    Data4(7) As Byte
End Type

Private Declare Function CreatePalette Lib "gdi32" (lpLogPalette As LOGPALETTE) As Long
Private Declare Function GetPaletteEntries Lib "gdi32" (ByVal hPalette As Long, ByVal wStartIndex As Long, ByVal wNumEntries As Long, lpPaletteEntries As PALETTEENTRY) As Long
         
Private Type PicBmp
    Size As Long
    Type As Long
    hBmp As Long
    hPal As Long
    Reserved As Long
End Type

'API FindFile / FileExists
'================================================================
Public Const INVALID_HANDLE_VALUE As Long = -1
Public Const MAX_PATH As Long = 260
Public Const FILE_ATTRIBUTE_ARCHIVE = &H20
Public Const FILE_ATTRIBUTE_COMPRESSED = &H800
Public Const FILE_ATTRIBUTE_DIRECTORY = &H10
Public Const FILE_ATTRIBUTE_HIDDEN = &H2
Public Const FILE_ATTRIBUTE_NORMAL = &H80
Public Const FILE_ATTRIBUTE_READONLY = &H1
Public Const FILE_ATTRIBUTE_SYSTEM = &H4
Public Const FILE_ATTRIBUTE_TEMPORARY = &H100

Public Type WIN32_FIND_DATA
    dwFileAttributes As Long
    ftCreationTime As FILETIME
    ftLastAccessTime As FILETIME
    ftLastWriteTime As FILETIME
    nFileSizeHigh As Long
    nFileSizeLow As Long
    dwReserved0 As Long
    dwReserved1 As Long
    cFileName As String * MAX_PATH
    cAlternate As String * 14
End Type

Public Declare Function FindFirstFile Lib "kernel32" Alias "FindFirstFileA" (ByVal lpFileName As String, lpFindFileData As WIN32_FIND_DATA) As Long
Public Declare Function FindNextFile Lib "kernel32" Alias "FindNextFileA" (ByVal hFindFile As Long, lpFindFileData As WIN32_FIND_DATA) As Long
Public Declare Function FindClose Lib "kernel32" (ByVal hFindFile As Long) As Long

Public Function LoadPCX(ByVal Filename As String) As StdPicture
    
    Dim nFreeFile As Integer
    Dim Header As PCXHeader
    
    nFreeFile = FreeFile

    Open Filename For Binary Lock Write As #nFreeFile
    
    Seek #nFreeFile, 1
    Get #nFreeFile, , Header
    Close #nFreeFile

    If Header.Bpp = 8 Then
        
        If Header.Planes = 1 Then
                    
            Set LoadPCX = Read8Bit(Filename)
                    
        ElseIf Header.Planes = 3 Then
        
            Set LoadPCX = Read24Bit(Filename)
        
        End If
    
    Else
        
        'error
        MsgBox "Invalid PCX format file.", vbApplicationModal + vbExclamation + vbOKOnly, "Loading PCX..."
        Set LoadPCX = Nothing
    
    End If

End Function


Private Function FileExists(ByVal Filename As String) As Boolean
Dim WFD As WIN32_FIND_DATA
Dim hFile As Long
Dim fn As String

If Right$(Filename, 1) <> Chr$(0) Then
    fn = Filename & Chr$(0)
Else
    fn = Filename
End If

hFile = FindFirstFile(Filename, WFD)
FileExists = (hFile <> INVALID_HANDLE_VALUE)

FindClose hFile
End Function


Private Function Read24Bit(ByRef Filename As String) As StdPicture

'==================================================
'PCX decoding functions for Visual Basic!
'
'(C) '99 Ryuunosuke (Dave Hng)
'Written with information from ZSoft.
'==================================================
'
'This one isn't so fast :)

'Init
'==================================================
    
    Dim nFreeFile As Integer
    Dim Header As PCXHeader
    Dim nWidth As Integer, nHeight As Integer
    Dim nLineSize As Integer
    Dim BITMAP() As RGBQUAD
    Dim cData() As Byte
    
    nFreeFile = FreeFile

'==================================================

'Read
'==================================================

    Open Filename For Binary Lock Write As #nFreeFile
    
    Seek #nFreeFile, 1
    Get #nFreeFile, , Header
    
    With Header
        nWidth = .XMax - .XMin + 1
        nHeight = .YMax - .YMin + 1
        nLineSize = .Planes * .BytesPerLine
    End With
    
    ReDim cData(LOF(nFreeFile) - Len(Header))
    
    Get #nFreeFile, , cData()
    
    Close #nFreeFile

'==================================================


'DWord align the X width, otherwise Windows reads it incorrectly!
'==================================================
    
    Dim nBitmapX As Long
    
    If (nWidth) Mod Len(nBitmapX) = 0 Then
        
        'If nWidth is divisible by four, then it's okay.
        nBitmapX = nWidth
    
    Else
        
        'int round down nWidth to the nearest multiple of 4
        'then add 3, because we're counting from 0.
        nBitmapX = (nWidth \ 4) * 4 + 3
    
    End If

'==================================================

'Resize
'==================================================

    ReDim BITMAP(nBitmapX * nHeight)

'==================================================


'Decode
'==================================================

    Dim nPos As Long, nRepeats As Long, nRepIter As Long
    Dim nScanPos As Long
    Dim cDecoded() As Byte
    
    ReDim cDecoded(nWidth * nHeight * 3&)
    
    For nPos = 0 To UBound(cData, 1)
    
        If (cData(nPos) And &HC0) = &HC0 Then
            
            'Encoded
            '==================================================
            nRepeats = cData(nPos) And &H3F
            
            For nRepIter = 0 To nRepeats - 1
                cDecoded(nScanPos + nRepIter) = cData(nPos + 1)
            Next nRepIter
        
            nPos = nPos + 1                 'Double incriment nPos
            nScanPos = nScanPos + nRepeats  'Indicate how many pixels added
            '==================================================
        
        Else
            
            'Not encoded
            '==================================================
            cDecoded(nScanPos) = cData(nPos)
            
            nScanPos = nScanPos + 1
            '==================================================
        End If
    
    Next nPos

    Erase cData

'==================================================


'Sort from planes into a bitmap
'==================================================

    Dim X As Long, Y As Long, nStartPos As Long
    
    For Y = 0 To nHeight - 1
        For X = 0 To nWidth - 1
        
            nStartPos = Y * Header.BytesPerLine * 3 + X
        
            With BITMAP(X + Y * nWidth)
                .rgbRed = cDecoded(nStartPos)
                .rgbGreen = cDecoded(nStartPos + Header.BytesPerLine)
                .rgbBlue = cDecoded(nStartPos + Header.BytesPerLine * 2)
            End With
        
        Next X
    Next Y
    
    Erase cDecoded

'==================================================


'Write Bmpinfo struct
'==================================================
    
    Dim bmpinfo As BITMAPINFO
       
    With bmpinfo.bmiHeader
        .biSize = Len(bmpinfo.bmiHeader)        'SizeOf Struct
        .biWidth = nWidth                       'Bitmap Width
        .biHeight = -nHeight                    'Bitmap Height, bitmap is top down.
        .biPlanes = 1                           'Single plane
        .biBitCount = 32                        '32 bit alignment
        .biCompression = BI_RGB                 'No Compression
        .biSizeImage = 0                        'Default
        .biXPelsPerMeter = 0                    'Default
        .biYPelsPerMeter = 0                    'Default
        .biClrUsed = 0                          'Default
        .biClrImportant = 0                     'Default
    End With

'==================================================

'Create the Bitmap
'==================================================
    
    Dim hBitmap As Long
    Dim hScreenDc As Long
    
    'Get the screen dc
    hScreenDc = GetDC(ByVal 0&)
    
    'Create the bitmap
    hBitmap = CreateDIBitmap(hScreenDc, bmpinfo.bmiHeader, CBM_INIT, BITMAP(0), bmpinfo, DIB_RGB_COLORS)
    
    Call ReleaseDC(ByVal 0&, hScreenDc)

'==================================================


'Create the StdPicture object
'==================================================
    
    Dim Pic As PicBmp
    Dim IPic As IPicture
    Dim IID_IDispatch As GUID
    
    With IID_IDispatch
        .Data1 = &H20400
        .Data4(0) = &HC0
        .Data4(7) = &H46
    End With
    
    With Pic
        .Size = Len(Pic)
        .Type = vbPicTypeBitmap
        .hBmp = hBitmap
        .hPal = 0
    End With
        
    OleCreatePictureIndirect Pic, IID_IDispatch, TrueAPI, IPic
    
    Set Read24Bit = IPic

'==================================================

'And we're done!
End Function

Private Function Read8Bit(ByRef Filename As String) As StdPicture

'==================================================
'PCX decoding functions for Visual Basic!
'
'(C) '99 Ryuunosuke (Dave Hng)
'Written with information from ZSoft.
'==================================================

'Init
'==================================================
    
    'Variables and arrays
    Dim nFreeFile As Integer
    Dim Header As PCXHeader
    Dim Data() As Byte
    Dim Palette(0 To 255) As RGBTriple
    Dim PaletteRGB(0 To 255) As Long
    Dim PalByte As Byte
    Dim result As Long
    
    'Constants used in decoding the PCX
    Const cCounterByte = &HC0
    '- Two most sig bits in a byte. If these two are high, then
    '  the byte's 6 other bits tell how many times to repeat data.
    
    Const cCounterMask = &H3F
    '- AND Mask for finding out how many repititions to do.
    
    Const cStartOfPalette = 12
    '- PCXs have this byte preceeding the 256 colour palette.
    
    'Find first free file handle.
    nFreeFile = FreeFile

'==================================================


'Read the data
'==================================================
    
    Open Filename For Binary Lock Write As #nFreeFile
    
    'Read the header
    Get #nFreeFile, , Header
    
    'Get data
    ReDim Data(LOF(nFreeFile) - Len(Header))
    Get #nFreeFile, , Data()
    
    'Get palette indication byte
    Seek #nFreeFile, LOF(nFreeFile) - 768
    Get #nFreeFile, , PalByte
    
    'Get Palette
    If PalByte = cStartOfPalette Then
        Seek #nFreeFile, LOF(nFreeFile) - 767
        Get #nFreeFile, , Palette()
    Else
        'error
        MsgBox "Invalid PCX format file.", vbApplicationModal + vbExclamation + vbOKOnly, "Loading PCX..."
        Exit Function
    End If
    
    Close #nFreeFile

'==================================================


'Get some info about the file
'==================================================
    
    Dim nXSize As Integer, nYSize As Integer
    Dim nLineSize As Integer
    
    nXSize = Header.XMax - Header.XMin + 1
    nYSize = Header.YMax - Header.YMin + 1
    
    nLineSize = Header.Planes * Header.BytesPerLine

'==================================================


'DWord align the X width, otherwise Windows reads it incorrectly!
'==================================================
    
    Dim nBitmapX As Long
    
    If (nLineSize) Mod Len(nBitmapX) = 0 Then
        
        'If nLineSize is divisible by four, then it's okay.
        nBitmapX = nLineSize - 1
    
    Else
        
        'int round down nLineSize to the nearest multiple of 4
        'then add 3, because we're counting from 0.
        nBitmapX = (nLineSize \ 4) * 4 + 3
    
    End If

'==================================================

'Allocate memory for the bitmap.
'==================================================

    Dim BITMAP() As Byte
    ReDim BITMAP(nBitmapX, nYSize - 1)

'==================================================

'Decompress the PCX!
'==================================================
    
    Dim nYCount As Integer
    Dim nIterCount As Integer
    Dim nBmpPos As Long
    Dim nNum As Integer
    Dim nBytePos As Long
    Dim nBmpCount As Long
    
    For nYCount = 0 To nYSize - 1
        
        Do Until nBmpCount >= nLineSize
            
            If (Data(nBytePos) And cCounterByte) = cCounterByte Then
                'If this is a 'counter' byte
                
                'Get the number of blocks to repeat
                nNum = Data(nBytePos) And cCounterMask
                
                'If extra data is present, ignore it.
                If nNum + nBmpCount > nLineSize Then
                    nNum = nXSize - nBmpCount
                End If
                
                'Repeat the blocks
                For nIterCount = 0 To nNum - 1
                    BITMAP(nBmpCount + nIterCount, nYCount) = Data(nBytePos + 1)
                Next nIterCount
                    
                nBmpCount = nBmpCount + nIterCount
                    
                'Increase the data count by 2
                nBytePos = nBytePos + 2
                
            Else
                'If this isn't a 'counter' byte
                
                'Put the data straight into the bitmap
                BITMAP(nBmpCount, nYCount) = Data(nBytePos)
            
                nBmpCount = nBmpCount + 1
                
                'And increase the data count by 1
                nBytePos = nBytePos + 1
            End If
        
        Loop
        
        nBmpCount = 0
        
    Next nYCount
    
    'Free data memory.
    Erase Data

'==================================================


'Construct the palette
'==================================================
    
    Dim nCount As Long
    
    'Create Palette entries, from triple to quad
    For nCount = 0 To 255
        PaletteRGB(nCount) = Palette(nCount).Red * &H10000 Or Palette(nCount).Green * &H100& Or Palette(nCount).Blue
    Next nCount
    
    'CopyMemory palette information over.
    Dim srcPtr As Long, dstPtr As Long
    
    For nCount = 0 To 255
        srcPtr = VarPtr(PaletteRGB(nCount))
        dstPtr = VarPtr(bmpinfo.bmiColors(nCount))
        
        RtlMoveMemory ByVal dstPtr, ByVal srcPtr, ByVal 4&
    Next nCount

'==================================================

'Write struct for windows
'==================================================
    
    With bmpinfo.bmiHeader
        .biBitCount = 8                     '8 bpp
        .biClrImportant = 0                 'All important colours
        .biClrUsed = 0                      'Windows can work this one out.
        .biCompression = BI_RGB             'No compression
        .biHeight = -nYSize                 'Pixels high, top down bitmap
        .biPlanes = 1                       '1 plane
        .biSize = Len(bmpinfo.bmiHeader)    'SizeOf struct
        .biSizeImage = 0                    'Not needed
        .biWidth = nLineSize                'Pixels wide
        .biXPelsPerMeter = 0                'Default
        .biYPelsPerMeter = 0                'Default
    End With

'==================================================


'Create the Bitmap
'==================================================
    
    Dim hScreenDc As Long
    
    'Get the screen dc
    hScreenDc = GetDC(ByVal 0&)
    
    'Create the bitmap
    hBitmap = CreateDIBitmap256(hScreenDc, bmpinfo.bmiHeader, CBM_INIT, BITMAP(0, 0), bmpinfo, DIB_RGB_COLORS)
    
    Call ReleaseDC(ByVal 0&, hScreenDc)

'==================================================


'Create the StdPicture object
'==================================================
    Dim Pic As PicBmp
    Dim IPic As IPicture
    Dim IID_IDispatch As GUID
    Dim hPal As Long
    Dim logPal As LOGPALETTE
    
    With logPal
        .palVersion = &H300
        .palNumEntries = 256
    End With
    
    'Needs another palette for some reason.
    For nCount = 0 To 255
        srcPtr = VarPtr(PaletteRGB(nCount))
        dstPtr = VarPtr(logPal.palPalEntry(nCount))
        
        RtlMoveMemory ByVal dstPtr, ByVal srcPtr, ByVal 4&
    Next nCount
    
    hPal = CreatePalette(logPal)
    
    With IID_IDispatch
        .Data1 = &H20400
        .Data4(0) = &HC0
        .Data4(7) = &H46
    End With
    
    With Pic
        .Size = Len(Pic)
        .Type = vbPicTypeBitmap
        .hBmp = hBitmap
        .hPal = hPal
    End With
        
    result = OleCreatePictureIndirect(Pic, IID_IDispatch, TrueAPI, IPic)
    
    Set Read8Bit = IPic

'==================================================

'And we're done!
End Function
