Attribute VB_Name = "basTheme"
'force variable declaration
Option Explicit


'color structure
Public Type ColorType
   Red As Integer
   Green As Integer
   Blue As Integer
End Type

'comment structure
Public Type CommentType
   Blank As Boolean
   Text As String
End Type

'gradient structure
Public Type GradientType
   Enabled As Boolean
   NumColors As Integer
   Ratio As Single
End Type

'background picture type
Public Type PictureType
   Filename As String
   Mode As Integer
End Type

'theme structure
Public Type ThemeType
   Name As String
   Enabled As Boolean
   NumComments As Integer
   Comments() As CommentType
   Color(0 To 8) As ColorType
   Gradient(0 To 1) As GradientType
   Special As String
   Picture As PictureType
End Type

'array
Public Theme() As ThemeType
Public LeadingComments() As CommentType

'indexs and counters
Public NumColor As Integer
Public NumLeadingComments As Integer
Public NumTheme As Integer
Public NumTotalThemes As Integer

'path to theme
Public ThemeFilePath As String


Public Sub CopyTheme(ByRef Source As ThemeType, ByRef Target As ThemeType)
   Dim a As Integer
   
   'copy source theme's data to target theme
   Target.Enabled = Source.Enabled
   Target.NumComments = Source.NumComments
   ReDim Target.Comments(0 To Target.NumComments) As CommentType
   For a = 0 To Target.NumComments
      Target.Comments(a).Blank = Source.Comments(a).Blank
      Target.Comments(a).Text = Source.Comments(a).Text
   Next a
   For a = 0 To 8
      Target.Color(a).Red = Source.Color(a).Red
      Target.Color(a).Green = Source.Color(a).Green
      Target.Color(a).Blue = Source.Color(a).Blue
   Next a
   For a = 0 To 1
      Target.Gradient(a).Enabled = Source.Gradient(a).Enabled
      Target.Gradient(a).NumColors = Source.Gradient(a).NumColors
      Target.Gradient(a).Ratio = Source.Gradient(a).Ratio
   Next a
   Target.Picture.Filename = Source.Picture.Filename
   Target.Picture.Mode = Target.Picture.Mode
   Target.Special = Source.Special
End Sub


Public Sub CreateTheme(ByRef Target As ThemeType)
   'set theme data to defaults
   Target.Enabled = False
   Target.NumComments = -1
   Target.Color(COLOR_BACKGROUND1).Red = 0
   Target.Color(COLOR_BACKGROUND1).Green = 0
   Target.Color(COLOR_BACKGROUND1).Blue = 0
   Target.Color(COLOR_BACKGROUND2).Red = 0
   Target.Color(COLOR_BACKGROUND2).Green = 0
   Target.Color(COLOR_BACKGROUND2).Blue = 0
   Target.Color(COLOR_BARS).Red = 31
   Target.Color(COLOR_BARS).Green = 31
   Target.Color(COLOR_BARS).Blue = 31
   Target.Color(COLOR_BORDER).Red = 63
   Target.Color(COLOR_BORDER).Green = 63
   Target.Color(COLOR_BORDER).Blue = 63
   Target.Color(COLOR_FILL).Red = 31
   Target.Color(COLOR_FILL).Green = 31
   Target.Color(COLOR_FILL).Blue = 31
   Target.Color(COLOR_HIGHLIGHT).Red = 0
   Target.Color(COLOR_HIGHLIGHT).Green = 0
   Target.Color(COLOR_HIGHLIGHT).Blue = 0
   Target.Color(COLOR_TEXT1).Red = 63
   Target.Color(COLOR_TEXT1).Green = 63
   Target.Color(COLOR_TEXT1).Blue = 63
   Target.Color(COLOR_TEXT2).Red = 63
   Target.Color(COLOR_TEXT2).Green = 63
   Target.Color(COLOR_TEXT2).Blue = 63
   Target.Color(COLOR_TEXT3).Red = 63
   Target.Color(COLOR_TEXT3).Green = 63
   Target.Color(COLOR_TEXT3).Blue = 63
   Target.Gradient(GRADIENT_MENUS).Enabled = True
   Target.Gradient(GRADIENT_MENUS).NumColors = 15
   Target.Gradient(GRADIENT_MENUS).Ratio = 1
   Target.Gradient(GRADIENT_TITLES).Enabled = True
   Target.Gradient(GRADIENT_TITLES).NumColors = 15
   Target.Gradient(GRADIENT_TITLES).Ratio = 1
   Target.Picture.Filename = ""
   Target.Picture.Mode = 0
   Target.Special = "none"
End Sub


Private Sub ParseColor(ByVal NumLine As Integer, ByRef Source As String, ByRef Color As Integer)
   Dim SubString As String
      
   SubString = ParseString(Source, 1, ",")
   If SubString = "" Then
      'error
      MsgBox "Missing color value at line " & NumLine & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
   Else
      If IsNumeric(SubString) = False Then
         'error
         MsgBox "Nonnumeric color value at line " & NumLine & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
      Else
         Color = CInt(SubString)
         If Color < 0 Then
            'error
            Color = 0
         Else
            If Color > 63 Then
               'error
               Color = 63
            End If
         End If
      End If
   End If
End Sub


Private Sub ParseGradient(ByVal NumLine As Integer, ByRef Source As String, ByRef Gradient As GradientType)
   Dim NumUnusedColors As Integer
   Dim Ratio As Single
   Dim SubString As String
      
   'enabled value
   SubString = ParseString(Source, 1, ",")
   If SubString = "" Then
      'error
      MsgBox "Missing gradient enabled value at line " & NumLine & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
   Else
      If (SubString <> "0") And (SubString <> "1") Then
         'error
         MsgBox "Gradient enabled value at line " & NumLine & " is not 0 or 1.", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
      Else
         Gradient.Enabled = CBool(-CInt(SubString))
      End If
   End If
               
   'number of colors
   SubString = ParseString(Source, 1, ",")
   If SubString = "" Then
      'error
      MsgBox "Missing number of unused gradient colors at line " & NumLine & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
   Else
      If IsNumeric(SubString) = False Then
         'error
         MsgBox "Nonnumeric value of unused gradient colors at line " & NumLine & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
      Else
         NumUnusedColors = CInt(SubString)
         If NumUnusedColors < 0 Then
            'error
            MsgBox "Number of unused colors at line " & NumLine & " is less than 0.", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
            NumUnusedColors = 0
         Else
            If NumUnusedColors > 15 Then
               'error
               MsgBox "Number of unused colors at line " & NumLine & " is greater than 15.", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
               NumUnusedColors = 15
            End If
         End If
         Gradient.NumColors = 15 - NumUnusedColors
      End If
   End If
   
   'ratio
   SubString = ParseString(Source, 1, " ")
   If SubString = "" Then
      'error
      MsgBox "Missing gradient ratio value at line " & NumLine & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
   Else
      If IsNumeric(SubString) = False Then
         'error
         MsgBox "Nonnumeric value of gradient ratio at line " & NumLine & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
      Else
         Ratio = CSng(Format$(SubString, "0.0"))
         If Ratio < 0 Then
            'error
            Ratio = 0
         Else
            If Ratio > 4 Then
               'error
               Ratio = 4
            End If
         End If
         Gradient.Ratio = Ratio
      End If
   End If
End Sub


Private Function ParseString(ByRef Source As String, ByVal NumTerminators As Integer, ParamArray Terminator() As Variant) As String
   Dim a As Integer
   Dim pos As Long
   
   'parse string until terminator sequence reached and remove substring and terminator from source string
   ParseString = ""
   For a = 1 To NumTerminators
      pos = InStr(1, Source, Terminator(a - 1), vbTextCompare)
      If pos > 0 Then
         ParseString = Left$(Source, pos - 1)
         Source = Trim(Right$(Source, Len(Source) - pos - Len(Terminator(a - 1)) + 1))
         ParseString = Trim(ParseString)
         Exit Function
      End If
   Next a
   
   'no terminators found
   ParseString = Source
   Source = ""
End Function


Public Sub ReadThemeFile(ByVal Path As String)
   Dim CommentedLine As Boolean
   Dim FileNumber As Integer
   Dim FileLine() As String
   Dim FileLineCount As Integer
   Dim OldString As String
   Dim SubString As String
   Dim a As Integer
   Dim b As Integer
   
   'open theme file
   FileNumber = FreeFile
   Open Path For Input Access Read Lock Write As #FileNumber
      
   'read file
   FileLineCount = -1
   Do Until EOF(FileNumber)
      FileLineCount = FileLineCount + 1
      ReDim Preserve FileLine(0 To FileLineCount) As String
      Line Input #FileNumber, FileLine(FileLineCount)
   Loop
   
   'close file
   Close #FileNumber
   
   'parse file into themes
   NumLeadingComments = -1
   NumTheme = -1
   For a = 0 To FileLineCount
      'remove leading and trailing spaces from line
      FileLine(a) = Trim(FileLine(a))
      
      'determine if comment
      If FileLine(a) = "" Then
         CommentedLine = True
      Else
         CommentedLine = False
         If Left$(FileLine(a), 1) = ";" Then
            CommentedLine = True
            FileLine(a) = Trim(Right$(FileLine(a), Len(FileLine(a)) - 1))
         End If
      End If
      
      'check for leading comments
      If (NumTheme = -1) And (CommentedLine = True) Then
         If FileLine(a) = "" Then
            NumLeadingComments = NumLeadingComments + 1
            ReDim Preserve LeadingComments(0 To NumLeadingComments) As CommentType
            LeadingComments(NumLeadingComments).Blank = True
         Else
            NumLeadingComments = NumLeadingComments + 1
            ReDim Preserve LeadingComments(0 To NumLeadingComments) As CommentType
            LeadingComments(NumLeadingComments).Blank = False
            LeadingComments(NumLeadingComments).Text = FileLine(a)
         End If
         GoTo NextLine
      End If
       
      'check for new theme name
      If Left$(FileLine(a), 1) = "[" Then
         'make sure there is a closing bracket
         If Right$(FileLine(a), 1) = "]" Then
            'make sure there is at least one character between brackets
            If Len(FileLine(a)) > 3 Then
               'make new theme
               NumTheme = NumTheme + 1
               ReDim Preserve Theme(0 To NumTheme) As ThemeType
               Theme(NumTheme).Name = Mid$(FileLine(a), 2, Len(FileLine(a)) - 2)
               Call CreateTheme(Theme(NumTheme))
               
               'set theme as enabled or disabled
               Theme(NumTheme).Enabled = Not CommentedLine
            Else
               'error
               MsgBox "Unnamed theme at line " & a + 1 & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
            End If
         Else
            'error
            MsgBox "No right bracket after theme name at line " & a + 1 & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
         End If
         GoTo NextLine
      End If
      
      'check variable name
      OldString = FileLine(a)
      SubString = LCase(ParseString(FileLine(a), 1, "="))
      Select Case SubString
         'background colors
         Case "background"
            With Theme(NumTheme).Color(COLOR_BACKGROUND1)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
            With Theme(NumTheme).Color(COLOR_BACKGROUND2)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
                        
         'background picture
         Case "background_pic"
            SubString = ParseString(FileLine(a), 2, ",", " ")
            If SubString = "" Then
               'error
                MsgBox "Missing background image filename at line " & (a + 1) & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
            Else
               If Dir(ThemeFilePath & SubString) = "" Then
                  'error
                  MsgBox "Image file """ & SubString & """ not found at line " & (a + 1) & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
               Else
                  Theme(NumTheme).Picture.Filename = SubString
               End If
            End If
            SubString = ParseString(FileLine(a), 1, " ")
            Select Case SubString
               Case "0", ""
                  Theme(NumTheme).Picture.Mode = 0
               Case "1"
                  Theme(NumTheme).Picture.Mode = 1
               Case "2"
                  Theme(NumTheme).Picture.Mode = 2
               Case "3"
                  Theme(NumTheme).Picture.Mode = 3
               Case Else
                  'error
                  MsgBox "Background image mode """ & SubString & """ is invalid at line " & (a + 1) & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
            End Select
         
         'menus and title bars color
         Case "bars"
            With Theme(NumTheme).Color(COLOR_BARS)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
            
         'menu and title bars gradient
         Case "bar_gradient"
            Call ParseGradient(a + 1, FileLine(a), Theme(NumTheme).Gradient(GRADIENT_TITLES))
            
         'border color
         Case "borders"
            With Theme(NumTheme).Color(COLOR_BORDER)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
            
         'window fill color
         Case "fill"
            With Theme(NumTheme).Color(COLOR_FILL)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
            
         'highlight color
         Case "highlight"
            With Theme(NumTheme).Color(COLOR_HIGHLIGHT)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
               
         'menu gradient
         Case "menu_gradient"
            Call ParseGradient(a + 1, FileLine(a), Theme(NumTheme).Gradient(GRADIENT_MENUS))
            
         'special effects
         Case "special"
            'store special effect value
            SubString = LCase(ParseString(FileLine(a), 1, " "))
            If SubString = "" Then
               'error
               MsgBox "Missing special effect value at line " & (a + 1) & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
            Else
               For b = 0 To frmTheme.lstSpecials.ListCount - 1
                  If frmTheme.lstSpecials.List(b) = SubString Then
                     Theme(NumTheme).Special = SubString
                     Exit For
                  End If
               Next b
            End If
                              
         'text colors
         Case "text"
            With Theme(NumTheme).Color(COLOR_TEXT1)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
            With Theme(NumTheme).Color(COLOR_TEXT2)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
            With Theme(NumTheme).Color(COLOR_TEXT3)
               Call ParseColor(a + 1, FileLine(a), .Red)
               Call ParseColor(a + 1, FileLine(a), .Green)
               Call ParseColor(a + 1, FileLine(a), .Blue)
            End With
            
         'no name
         Case ""
            If CommentedLine = True Then
               'blank comment
               If NumTheme >= 0 Then
                  Theme(NumTheme).NumComments = Theme(NumTheme).NumComments + 1
                  ReDim Preserve Theme(NumTheme).Comments(0 To Theme(NumTheme).NumComments) As CommentType
                  Theme(NumTheme).Comments(Theme(NumTheme).NumComments).Blank = True
               Else
                  NumLeadingComments = NumLeadingComments + 1
                  ReDim Preserve LeadingComments(0 To NumLeadingComments) As CommentType
                  LeadingComments(NumLeadingComments).Blank = True
               End If
            Else
               'error
               MsgBox "No named variable at line " & a + 1 & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
            End If
            GoTo NextLine
            
         'unknown variable
         Case Else
            If CommentedLine = True Then
               'remove second comment character if exists
               If Left$(OldString, 1) = ";" Then
                  OldString = Trim(Right$(OldString, Len(OldString) - 1))
               End If
               
               'comment
               If NumTheme >= 0 Then
                  Theme(NumTheme).NumComments = Theme(NumTheme).NumComments + 1
                  ReDim Preserve Theme(NumTheme).Comments(0 To Theme(NumTheme).NumComments) As CommentType
                  Theme(NumTheme).Comments(Theme(NumTheme).NumComments).Blank = False
                  Theme(NumTheme).Comments(Theme(NumTheme).NumComments).Text = OldString
               Else
                  NumLeadingComments = NumLeadingComments + 1
                  ReDim Preserve LeadingComments(0 To NumLeadingComments) As CommentType
                  LeadingComments(NumLeadingComments).Blank = False
                  LeadingComments(NumLeadingComments).Text = OldString
               End If
            Else
               'error
               MsgBox "Unknown variable """ & SubString & """ at line " & a + 1 & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
            End If
            GoTo NextLine
      End Select
         
      'check for extra characters
      If FileLine(a) <> "" Then
         'error
         MsgBox "Extraneous characters at line " & a + 1 & ".", vbApplicationModal + vbExclamation + vbOKOnly, "Processing Theme..."
      End If

NextLine:
   Next a
   
   'get number of themes
   NumTotalThemes = NumTheme + 1
   
   'add a default theme if no themes
   If NumTotalThemes = 0 Then
      NumTotalThemes = 1
      ReDim Theme(0 To 0) As ThemeType
      Theme(NumTheme).Name = "Default"
      Call CreateTheme(Theme(0))
      Theme(NumTheme).Enabled = True
   End If
End Sub


Public Sub SaveThemeFile(ByVal Path As String)
   Dim LineLeader As String
   Dim FileNumber As Integer
   Dim a As Integer
   Dim b As Integer
   
   'rewrite theme file
   FileNumber = FreeFile
   Open Path For Output Access Write Lock Read Write As #FileNumber
   
   'write leading comments
   For a = 0 To NumLeadingComments
      If LeadingComments(a).Blank = True Then
         Print #FileNumber, ""
      Else
         Print #FileNumber, "; " & LeadingComments(a).Text
      End If
   Next a
   
   'write themes
   For a = 0 To NumTotalThemes - 1
      'set line leader
      If Theme(a).Enabled = False Then
         LineLeader = ";"
      Else
         LineLeader = ""
      End If
      
      'name
      Print #FileNumber, LineLeader & "[" & Theme(a).Name & "]"
      
      'comments
      For b = 0 To Theme(a).NumComments - 1
         If Theme(a).Comments(b).Blank = True Then
            Print #FileNumber, LineLeader
         Else
            Print #FileNumber, LineLeader & "; " & Theme(a).Comments(b).Text
         End If
      Next b
      
      'colors and background image
      Print #FileNumber, LineLeader & "background = " & Theme(a).Color(COLOR_BACKGROUND1).Red & ", " & Theme(a).Color(COLOR_BACKGROUND1).Green & ", " & Theme(a).Color(COLOR_BACKGROUND1).Blue & ", " & Theme(a).Color(COLOR_BACKGROUND2).Red & ", " & Theme(a).Color(COLOR_BACKGROUND2).Green & ", " & Theme(a).Color(COLOR_BACKGROUND2).Blue
      If Theme(a).Picture.Filename <> "" Then
         If Theme(a).Picture.Mode = 0 Then
            Print #FileNumber, LineLeader & "background_pic = " & Theme(a).Picture.Filename
         Else
            Print #FileNumber, LineLeader & "background_pic = " & Theme(a).Picture.Filename & ", " & Theme(a).Picture.Mode
         End If
      End If
      Print #FileNumber, LineLeader & "text = " & Theme(a).Color(COLOR_TEXT1).Red & ", " & Theme(a).Color(COLOR_TEXT1).Green & ", " & Theme(a).Color(COLOR_TEXT1).Blue & ", " & Theme(a).Color(COLOR_TEXT2).Red & ", " & Theme(a).Color(COLOR_TEXT2).Green & ", " & Theme(a).Color(COLOR_TEXT2).Blue & ", " & Theme(a).Color(COLOR_TEXT3).Red & ", " & Theme(a).Color(COLOR_TEXT3).Green & ", " & Theme(a).Color(COLOR_TEXT3).Blue
      Print #FileNumber, LineLeader & "fill = " & Theme(a).Color(COLOR_FILL).Red & ", " & Theme(a).Color(COLOR_FILL).Green & ", " & Theme(a).Color(COLOR_FILL).Blue
      Print #FileNumber, LineLeader & "borders = " & Theme(a).Color(COLOR_BORDER).Red & ", " & Theme(a).Color(COLOR_BORDER).Green & ", " & Theme(a).Color(COLOR_BORDER).Blue
      Print #FileNumber, LineLeader & "highlight = " & Theme(a).Color(COLOR_HIGHLIGHT).Red & ", " & Theme(a).Color(COLOR_HIGHLIGHT).Green & ", " & Theme(a).Color(COLOR_HIGHLIGHT).Blue
      Print #FileNumber, LineLeader & "bars = " & Theme(a).Color(COLOR_BARS).Red & ", " & Theme(a).Color(COLOR_BARS).Green & ", " & Theme(a).Color(COLOR_BARS).Blue
      
      'gradients
      Print #FileNumber, LineLeader & "bar_gradient = " & -CInt(Theme(a).Gradient(GRADIENT_TITLES).Enabled) & ", " & 15 - Theme(a).Gradient(GRADIENT_TITLES).NumColors & ", " & Format$(Theme(a).Gradient(GRADIENT_TITLES).Ratio, "0.0")
      Print #FileNumber, LineLeader & "menu_gradient = " & -CInt(Theme(a).Gradient(GRADIENT_MENUS).Enabled) & ", " & 15 - Theme(a).Gradient(GRADIENT_MENUS).NumColors & ", " & Format$(Theme(a).Gradient(GRADIENT_MENUS).Ratio, "0.0")
      
      'special effect
      If Theme(a).Special <> "none" Then
         Print #FileNumber, LineLeader & "special = " & Theme(a).Special
      End If
      
      'blank line
      If a <> NumTotalThemes - 1 Then
         Print #FileNumber, ""
      End If
   Next a
   
   'close file
   Close #FileNumber
End Sub
